package webovyformular.dat_typy;

import java.util.*;

/**
 * <p>Description: </p> Tda slou pro reprezentaci bezkontextovch gramatik
 * a pro prci s nimi.
 *
 * <p>Copyright: Copyright (c) 2005</p>
 *
 * @author Michal Hanzelka
 * @version 1.0
 */


public class Gramatika {
  private HashSet<Znak> neterminaly;
  private HashSet<Znak> terminaly;
  private HashMap<Znak, HashSet<ArrayList<Znak>>> pravidla;
  private Znak pocatek;

  /**
   * Vytvo se nov gramatika, kter bude obsahovat terminly, neterminly,
   * pravidla a poten neterminl zadan na vstupu.
   *
   * @param neterminaly HashSet - mnoina terminl
   * @param terminaly HashSet - mnoina neterminl
   * @param pravidla HashMap - mnoina pravidel
   * @param pocatek Znak - poten neterminl
   */
  public Gramatika(HashSet<Znak> neterminaly,
                   HashSet<Znak> terminaly,
                   HashMap<Znak, HashSet<ArrayList<Znak>>> pravidla,
                   Znak pocatek) {
    this.neterminaly = neterminaly;
    this.terminaly = terminaly;
    this.pravidla = pravidla;
    this.pocatek = pocatek;
  }

  /**
   * Metoda equals porovn gramatiku s objektem obj a vrt true, pokud je
   * objekt gramatika a obsahuje stejn mnoiny neterminl, terminl a
   * pravidel a ma i stejn poten neterminl.
   *
   * @param obj Object - objekt
   * @return boolean - vrac true, kdy jsou gramatika a objekt toton
   */
  public boolean equals(Object obj) {
    if (obj == this) {
      return true;
    }
    if ( (obj instanceof Gramatika) == false) {
      return false;
    }
    Gramatika g = (Gramatika) obj;

    if (!this.pocatek.equals(g.pocatek)) {
      return false;
    }
    if (!this.neterminaly.equals(g.neterminaly)) {
      return false;
    }
    if (!this.terminaly.equals(g.terminaly)) {
      return false;
    }

    return (this.pravidla.equals(g.pravidla));
  }

  /**
   * Metoda hashCode vrac hodnotu hesovaciho kodu pro danou gramatiku. Hash
   * code se vypot jako souet hash code vech neterminl, terminl,
   * pravidel a potenho neterminlu.
   *
   * @return int - hodnota hash code pro danou gramatiku
   */
  public int hashCode() {
    int h = 0;

    for (Znak aktNeterm : neterminaly) {
      h += aktNeterm.hashCode();
    }

    for (Znak aktTerm : terminaly) {
      h += aktTerm.hashCode();
    }

    for (Iterator itP = this.pravidla.entrySet().iterator(); itP.hasNext(); ) {
      Map.Entry dvojice = (Map.Entry) itP.next();
      h += ( (Znak) dvojice.getKey()).hashCode() +
          ( (HashSet<ArrayList<Znak>>) dvojice.getValue()).hashCode();
    }

    h += this.pocatek.hashCode();
    return h;
  }

  /**
   * Pevede gramatiku na <b>String</b> tak, e vsledn etzec obsahuje
   * nsledujc: <br/><br/>
   *  <i>   "Mnoina neterminl: {" + n + "} \n" + <br/>
   *     "Mnoina terminl: {" + t + "} \n" +   <br/>
   *     "Mnoina pravidel: {\n" + p + "} \n" +  <br/>
   *     "Poteni neterminl: " + pocatek; </i><br/>
   *<br/>
   * kde <i>n</i> jsou jmna neterminl oddlena mezerami, <br/>
   *     <i>t</i> jsou jmna terminl oddlena mezerami,   <br/>
   *     <i>p</i> jsou pravidla gramatiky zobrazen tak, e na kadm dku je
   *       neterminl nsledovn znaky <i>" -> "</i> a za nimi etzce, na kter
   *       se me dan neterminl pepsat, oddlen znakem <i>"|"</i>. Pravidla
   *       potenho neterminlu jsou uvedena jako prvn.
   *
   * @return String - String reprezentujc danou gramatiku
   */
  public String toString() {
    String s = "";

    String n = "";
    for (Iterator itn = this.neterminaly.iterator(); itn.hasNext(); ) {
      n += (Znak) itn.next();
      if (itn.hasNext()) {
        n += ", ";
      }
    }

    String t = "";
    for (Iterator itt = this.terminaly.iterator(); itt.hasNext(); ) {
      t += (Znak) itt.next();
      if (itt.hasNext()) {
        t += ", ";
      }
    }

    String p = "";
    HashSet<ArrayList<Znak>> aktPravidla;

    if (this.pravidla.containsKey(this.pocatek)) { //pravidla pocatecniho neterminlu se vypisuji prvni
      p += this.pocatek + " -> ";
      aktPravidla = this.pravidla.get(this.pocatek);
      for (Iterator itP = aktPravidla.iterator(); itP.hasNext(); ) {
        p += Retezec.toString( (ArrayList<Znak>) itP.next());
        if (itP.hasNext()) {
          p += " | ";
        }
      }
      p += "\n";
    }
    for (Iterator itN = this.pravidla.keySet().iterator(); itN.hasNext(); ) {
      Znak aktNeterm = (Znak) itN.next();
      if (!aktNeterm.equals(this.pocatek)) {
        p += aktNeterm + " -> ";
        aktPravidla = (HashSet<ArrayList<Znak>>)this.pravidla.get(aktNeterm);

        for (Iterator itP = aktPravidla.iterator(); itP.hasNext(); ) {
          p += Retezec.toString( (ArrayList<Znak>) itP.next());

          if (itP.hasNext()) {
            p += " | ";
          }
        }
        if (itN.hasNext()) {
          p += "\n";
        }
      }
    }

    s += "Mnoina neterminl: {" + n + "}\n" +
         "Mnoina terminl: {" + t + "}\n" +
         "Mnoina pravidel: {\n" + p + "}\n" +
         "Poteni neterminl: " + this.pocatek;
    return s;
  }

  /**
   * Vrac mnoinu neterminl.
   * @return HashSet - mnoina neterminl
   */
  public HashSet<Znak> getNeterminaly() {
    return this.neterminaly;
  }

  /**
   * Vrac mnoinu terminl.
   * @return HashSet - mnoina terminl
   */
  public HashSet<Znak> getTerminaly() {
    return this.terminaly;
  }

  /**
   * Vrac mnoinu pravidel.
   * @return HashMap - mnoina pravidel
   */
  public HashMap<Znak, HashSet<ArrayList<Znak>>> getPravidla() {
    return this.pravidla;
  }

  /**
   * Vrac poten neterminl.
   * @return Znak - poten neterminl
   */
  public Znak getPocatek() {
    return this.pocatek;
  }

  /**
   * Metoda z dan gramatiky odstran ty neterminly, a pravidla ve kterch se
   * vyskytuj, ze kterch nelze odvodit terminln etzec. Na vstup je
   * vrcena upraven gramatika. Pvodn gramatika se nemn.
   * @return Gramatika - gramatika, kter neobsahuje neterminly, ze kterch
   * nelze odvodit terminln etzec
   */
  private Gramatika redukujNeterminaly() {

    //  int j = 0;

    HashSet<Znak> dosazNetPred = new HashSet<Znak> (); //dosazitelne neterminaly v predchozim cyklu
    HashSet<Znak> dosazNetAkt = new HashSet<Znak> (); //dosazitelne neterminaly v aktualni cyklu

    /* mnozina zespoda nedosazitelnych neterminalu, na pocatku vsechny */
    HashSet<Znak> nedosazNeterm = new HashSet<Znak> ();
    nedosazNeterm.addAll(pravidla.keySet());

    boolean pridejDoN; //pridej do dosazitelnych neterminalu

    do {
      dosazNetPred.addAll(dosazNetAkt);
      nedosazNeterm.removeAll(dosazNetPred);

      for (Znak aktNeterm : nedosazNeterm) { // prochazim jen ty neterminaly,
                                             // ktere jeste nejsou dosazitelne
        HashSet<ArrayList<Znak>> aktNetermPravidla = pravidla.get(aktNeterm);
        if (aktNetermPravidla != null) {

          pridejDoN = true;
          for (ArrayList<Znak> aktPravidlo : aktNetermPravidla) { //prochazim aktNetermpravidla
            pridejDoN = true; //predpokladam, ze aktPravidlo je prijatelne

            for (Znak aktZnak : aktPravidlo) { //prochazim aktPravidlo
              if ( (aktZnak.isNeterminal()) &&
                  (!dosazNetAkt.contains(aktZnak))) { //neterminal neni dosazitelny
                pridejDoN = false;
                break; //pravidlo neni treba dal zkoumat -> neni prijatelne
              }
            }

            if (pridejDoN) { //aktPravidlo obsahuje jen prijatelne znaky
              break; //netreba zkoumat dalsi pravidla -> aktNeterm je prijat
            }
          }

          if (pridejDoN) { //aktNeterm ma alespon jedno prijatelne pravidlo
            dosazNetAkt.add(aktNeterm);
          }
        }
      }

    }
    while (!dosazNetAkt.equals(dosazNetPred)); //mnoziny jsou stejne => konec

    /* odtraneni pravidel obsahujici zdola nedosazitelne neterminaly */

    HashMap<Znak, HashSet<ArrayList<Znak>>> novaPrav;
    novaPrav = new HashMap<Znak, HashSet<ArrayList<Znak>>> ();
    boolean prijatelne; //pravidlo je prijatelne, kdyz obsahuje pouze
                        //dosazitelne neterminaly nebo terminaly

    for (Znak aktNeterm : dosazNetAkt) {

      HashSet<ArrayList<Znak>> aktNetermPravidla = pravidla.get(aktNeterm);
      HashSet<ArrayList<Znak>> prijatelnaPrav = new HashSet<ArrayList<Znak>> ();

      for (ArrayList<Znak> aktPrav : aktNetermPravidla) {
        prijatelne = true;

        for (Znak aktZnak : aktPrav) {
          if ( (aktZnak.isNeterminal()) &&        //neterminal je neprijatelny
              (!dosazNetAkt.contains(aktZnak))) {
            prijatelne = false;          //pravidlo je neprijatelne
            break;                       //netreba zkoumat dalsi znaky pravidla
          }
        }
        if (prijatelne) {
          prijatelnaPrav.add(aktPrav); //pridani pravidla mezi prijatelne
        }
      }
      novaPrav.put(aktNeterm, prijatelnaPrav); //pridana jen prijatelna pravidla
    }

    Gramatika gramatika = new Gramatika(dosazNetAkt, this.terminaly, novaPrav,
                                        this.pocatek);

    return gramatika;
  }

  /**
   * Metoda odstran z gramatiky nepouiteln symboly a tuto redukovanou
   * gramatiku vrac na vstup. Pvodn gramatika zstv nezmnna.
   * @return Gramatika - gramatika bez nepouitelnch symbol
   */
  public Gramatika redukuj() {

    int j = 0;

    Gramatika g = this.redukujNeterminaly(); //odstrani zespoda nedosazitelne
                                             //neterminaly a pravidla, ve
                                             //kterych se vyskytuji


    /* mnozina z pocatku dosazitelnych pravidel */
    HashMap<Znak, HashSet<ArrayList<Znak>>> dosazPrav;
    dosazPrav = new HashMap<Znak, HashSet<ArrayList<Znak>>> ();

    HashSet<Znak> dosazNetPred = new HashSet<Znak> (); //dosazitelne neterm. v predchozim cyklu
    HashSet<Znak> dosazNetAkt = new HashSet<Znak> (); //dosazitelne neterm. v aktualni cyklu

    HashSet<Znak> dosazTermPred = new HashSet<Znak> (); //dosazitelne term. v predchozim cyklu
    HashSet<Znak> dosazTermAkt = new HashSet<Znak> (); //dosazitelne term. v aktualni cyklu

    dosazNetAkt.add(pocatek); //inicializace... dosazitelny je jen pocatek

    do {
      dosazTermPred.addAll(dosazTermAkt);
      dosazNetPred.addAll(dosazNetAkt);

      /* pruchod dosazitelne neterminaly */
      for (Znak aktNeterm : dosazNetPred) {

        HashSet<ArrayList<Znak>> aktNetermPravidla = g.pravidla.get(aktNeterm);

        if (aktNetermPravidla != null) { //pro aktNeterm existuji pravidla
          dosazPrav.put(aktNeterm, aktNetermPravidla); //dosazitelna pravidla

          for (ArrayList<Znak> aktPravidlo : aktNetermPravidla) {
            for (j = 0; j < aktPravidlo.size(); j++) {
              if (aktPravidlo.get(j).isNeterminal()) {
                dosazNetAkt.add(aktPravidlo.get(j));
              }
              else {
                if (aktPravidlo.get(j).isTerminal()) {
                  dosazTermAkt.add(aktPravidlo.get(j));
                }
              }
            } //for
          } //for
        } //if
      } //for

    }
    while (! (dosazTermAkt.equals(dosazTermPred) &&
              dosazNetAkt.equals(dosazNetPred))); //mnoziny se rovnaji => konec

    Gramatika redukG = new Gramatika(dosazNetAkt, dosazTermAkt, dosazPrav,
                                     pocatek);
    return redukG;
  }

  /**
   * Metoda rozhoduje, zda je zadan gramatika SLL_k pro zadan k. Metoda vrac
   * true prv tehdy, kdy je zadan gramatika SLL_k.Podminky pro to, aby
   * gramatika byla SLL_k: <br/>
   * pro vechny neterminly A z mnoiny neterminl a pravidla ve tvaru <br/>
   * <i> A->alfa, A->beta, kde alfa != beta </i>, plat:         <br/>
   * <i> First_k(alfa.Follow_k(A)) "prnik s" First_k(beta.Follow_k(A)) =
   * "przdn mnoina" </i>
   *
   * @param k int - paramter k
   * @return boolean - true, kdy je dan gamatika SLL_k
   * @throws NekladnyParamException  kdy parametr nen kladn slo
   */
  public boolean isSll(int k) throws NekladnyParamException {

    for (Znak aktNeterm : this.neterminaly) {
      HashSet<ArrayList<Znak>> aktNetermPravidla = this.pravidla.get(aktNeterm);

      for (ArrayList<Znak> alfa : aktNetermPravidla) {

        /* pruchod pres vsechna znaky aktNetermPravidla */
        for (ArrayList<Znak> beta : aktNetermPravidla) {
          if (!alfa.equals(beta)) { //podminka  alfa != beta

            /* alfaFO = alfa.Follow_k(aktNeter) */
            HashSet<ArrayList<Znak>> alfaFO =
                Retezec.zretez(alfa,
                               Retezec.follow(aktNeterm, this, k));
            /* betaFO = beta.Follow_k(aktNeter) */
            HashSet<ArrayList<Znak>> betaFO =
                Retezec.zretez(beta,
                               Retezec.follow(aktNeterm, this, k));

            /* alfaFI = First_k(alfaFO) */
            HashSet<ArrayList<Znak>> alfaFI = new HashSet<ArrayList<Znak>> ();
            for (ArrayList<Znak> aktAlfa1FO : alfaFO) {
              alfaFI.addAll(Retezec.first(aktAlfa1FO, this.pravidla, k));
            }

            /* betaFI = First_k(betaFO) */
            HashSet<ArrayList<Znak>> betaFI = new HashSet<ArrayList<Znak>> ();
            for (ArrayList<Znak> aktBetaFO : betaFO) {
              betaFI.addAll(Retezec.first(aktBetaFO, this.pravidla, k));
            }

            alfaFI.retainAll(betaFI); //alfaFI = alfaFI prunik betaFI
            if (!alfaFI.isEmpty()) {
              return false; //prunik je nenulovy -> gramatika neni SLL_k
            }
          } //if

        } //for
      } //for
    } //for
    return true;
  }

}
